#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Unit tests, to be run using pytest."""


__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020-2025 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import kineticstoolkit as ktk
import numpy as np


def test_read_n3d():
    """Regression test."""
    markers = ktk.ext.n3d.read_n3d(
        ktk.ext.root_folder + "/data/n3d_sample_optotrak.n3d"
    )

    tol = 1e-4
    assert np.abs(np.sum(markers.data["Marker0"]) - 172.3365) < tol
    assert np.abs(np.sum(markers.data["Marker40"]) + 45.3753) < tol
    assert markers.time_info["Unit"] == "s"
    assert markers.data_info["Marker40"]["Unit"] == "m"

    labels = [
        "Probe1",
        "Probe2",
        "Probe3",
        "Probe4",
        "Probe5",
        "Probe6",
        "FRArrD",
        "FRArrG",
        "FRav",
        "ScapulaG1",
        "ScapulaG2",
        "ScapulaG3",
        "ScapulaD1",
        "ScapulaD2",
        "ScapulaD3",
        "Tete1",
        "Tete2",
        "Tete3",
        "Sternum",
        "BrasG1",
        "BrasG2",
        "BrasG3",
        "EpicondyleLatG",
        "AvBrasG1",
        "AvBrasG2",
        "AvBrasG3",
        "NAG",
        "GantG1",
        "GantG2",
        "GantG3",
        "BrasD1",
        "BrasD2",
        "BrasD3",
        "EpicondyleLatD",
        "AvBrasD1",
        "AvBrasD2",
        "AvBrasD3",
        "NAD",
        "GantD1",
        "GantD2",
        "GantD3",
    ]

    markers = ktk.ext.n3d.read_n3d(
        ktk.ext.root_folder + "/data/n3d_sample_optotrak.n3d",
        labels=labels,
    )

    assert np.abs(np.sum(markers.data["Probe1"]) - 172.3365) < tol
    assert np.abs(np.sum(markers.data["GantD3"]) + 45.3753) < tol
    assert markers.time_info["Unit"] == "s"
    assert markers.data_info["GantD3"]["Unit"] == "m"


if __name__ == "__main__":
    import pytest

    pytest.main([__file__])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Read NDI Optotrak N3D files.
"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2022 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


from kineticstoolkit import TimeSeries
import struct
import numpy as np
from typing import Sequence


def read_n3d(filename: str, labels: Sequence[str] = []) -> TimeSeries:
    """
    Read markers from an NDI N3D file.

    The markers positions are returned in a TimeSeries where each marker
    corresponds to a data key. Each marker position is expressed in this form:

    array([[x0, y0, z0, 1.], [x1, y1, z1, 1.], [x2, y2, z2, 1.], ...])

    Parameters
    ----------
    filename : str
        Path of the N3D file.
    labels : list of str (optional)
        Marker names

    Returns
    -------
    TimeSeries

    """
    with open(filename, "rb") as fid:
        _ = fid.read(1)  # 32
        n_markers = struct.unpack("h", fid.read(2))[0]
        n_data_per_marker = struct.unpack("h", fid.read(2))[0]
        n_columns = n_markers * n_data_per_marker

        n_frames = struct.unpack("i", fid.read(4))[0]

        collection_frame_frequency = struct.unpack("f", fid.read(4))[0]
        user_comments = struct.unpack("60s", fid.read(60))[0]
        system_comments = struct.unpack("60s", fid.read(60))[0]
        file_description = struct.unpack("30s", fid.read(30))[0]
        cutoff_filter_frequency = struct.unpack("h", fid.read(2))[0]
        time_of_collection = struct.unpack("8s", fid.read(8))[0]
        _ = fid.read(2)
        date_of_collection = struct.unpack("8s", fid.read(8))[0]
        extended_header = struct.unpack("73s", fid.read(73))[0]

        # Read the rest and put it in an array
        ndi_array = np.ones((n_frames, n_columns)) * np.nan

        for i_frame in range(n_frames):
            for i_column in range(n_columns):
                data = struct.unpack("f", fid.read(4))[0]
                if data < -1e25:  # technically, it is -3.697314e+28
                    data = np.nan
                ndi_array[i_frame, i_column] = data

        # Conversion from mm to meters
        ndi_array /= 1000

        # Transformation to a TimeSeries
        ts = TimeSeries(
            time=np.linspace(
                0, n_frames / collection_frame_frequency, n_frames
            )
        )

        for i_marker in range(n_markers):
            if labels != []:
                label = labels[i_marker]
            else:
                label = f"Marker{i_marker}"

            ts.data[label] = np.block(
                [
                    [
                        ndi_array[:, 3 * i_marker : 3 * i_marker + 3],
                        np.ones((n_frames, 1)),
                    ]
                ]
            )
            ts = ts.add_data_info(label, "Unit", "m")

    return ts



if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2025 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides objects to find events from videos.

Warning
-------
This is currently in development, the whole interface could change at any time.

This module requires OpenCV which is not a dependency (yet) of Kinetics
Toolkit.

Installing OpenCV
-----------------
- Using conda: `conda install -c conda-forge opencv`
- Using pip: `pip install opencv-python`

"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2025 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
from matplotlib import animation, widgets
from kineticstoolkit import TimeSeries
from kineticstoolkit.exceptions import warn_once
from kineticstoolkit.typing_ import check_param
import time


WINDOW_PLACEMENT = {"top": 50, "right": 0}


def read_video(
    filename: str,
    original_fps: float | None = None,
    width: int | None = None,
    height: int | None = None,
    downsample: int = 1,
    max_memory: int = 1000,
) -> TimeSeries:
    """
    Read a video file as a TimeSeries.

    Warning
    -------
    This feature is experimental and will probably change in the future.

    Reads a video file using opencv, and returns a TimeSeries with one data
    key ("Video") that contains the video as an NxHxWx3 array where N is the
    number of images in the video, H is the video height, W is the video
    width, and 3 are the RGB channels.

    Parameters
    ----------
    filename
        The name of the video file
    original_fps
        Optional. Specifies the sampling rate of the video, in frames per
        second. Useful in cases where a video reports the wrong sampling rate,
        particularly in slow motion videos.
    width
        Optional. Scale down the video to a given width in pixels.
        Aspect ratio is maintained unless both width and height are set.
    height
        Optional. Scale down the video to a given height in pixels.
        Aspect ratio is maintained unless both width and height are set.
    downsample
        Optional. Downsample the video by a given integer. For instance, a
        240 FPS video with a downsample parameter of 2 will results in a
        120 FPS TimeSeries. Default is 1.
    max_memory
        Prevent the output TimeSeries from exceeding a maximal RAM size, in
        Mb. Default is 1000.

    Returns
    -------
    TimeSeries

    """
    warn_once(
        "Function read_video is experimental and will probably change in the "
        "future."
    )

    # Type check
    check_param("filename", filename, str)
    if original_fps is not None:
        check_param("original_fps", original_fps, float)
    check_param("width", width, int)
    check_param("height", height, int)
    check_param("downsample", downsample, int)
    check_param("max_memory", max_memory, int)

    # Import cv2 now, not on the top because opencv is not installed by
    # default.
    import cv2  # noqa

    video = cv2.VideoCapture(filename)

    video_length = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
    video_width = int(video.get(cv2.CAP_PROP_FRAME_WIDTH))
    video_height = int(video.get(cv2.CAP_PROP_FRAME_HEIGHT))

    if original_fps is None:
        video_fps = video.get(cv2.CAP_PROP_FPS)
    else:
        video_fps = original_fps

    new_video_length = video_length // downsample
    new_video_fps = video_fps / downsample

    if (width is not None) and (height is None) and (width < video_width):
        new_video_width = int(width)
        new_video_height = int(video_height * (width / video_width))
    elif (width is None) and (height is not None) and (height < video_height):
        new_video_height = int(height)
        new_video_width = int(video_width * (height / video_height))
    elif (
        (width is not None)
        and (height is not None)
        and ((height < video_height) | (width < video_width))
    ):
        new_video_height = int(height)
        new_video_width = int(width)
    else:
        new_video_width = int(video_width)
        new_video_height = int(video_height)

    estimated_memory_size = int(
        new_video_length * new_video_width * new_video_height * 3 / 1e6
    )
    if estimated_memory_size > max_memory:
        raise ValueError(
            "The resulting TimeSeries' RAM usage would be about "
            f"{estimated_memory_size} Mb, which exceed "
            f"the maximal RAM usage of {max_memory} Mb "
            "defined by the max_memory parameter. You can either increase "
            "this value or reduce the memory required by the video by "
            "resizing or downsampling it using the according "
            "read_video function parameters."
        )

    frames = np.empty(
        (new_video_length, new_video_height, new_video_width, 3),
        dtype=np.uint8,
    )

    for i in tqdm(range(video_length)):
        read, frame = video.read()

        if not read:
            break

        if i % downsample != 0:
            continue

        if new_video_width != video_width:
            frame = cv2.resize(
                frame,
                (new_video_width, new_video_height),
                interpolation=cv2.INTER_LINEAR,
            )
        frames[i // downsample] = frame[:, :, -1::-1]  # BGR to RGB

    return TimeSeries(
        time=np.arange(new_video_length) / new_video_fps,
        data={"Video": frames},
    )


def ui_edit_events(
    ts_video: TimeSeries,
    ts_data: TimeSeries,
    video_key: str = "Video",
    data_keys: str | list[str] = [],
    event_source: str = "merge",
    in_place: bool = False,
) -> list[TimeSeries]:
    """
    Use an interactive interface to edit time and events in a Video TimeSeries.

    Warning
    -------
    This feature is experimental and will probably change in the future.

    Parameters
    ----------
    ts_video
        The TimeSeries that contains the video, normally obtained using
        the read_video() function.
    ts_data
        Optional. TimeSeries.
    video_key
        Optional. The data key in ts_video that contains the video. Default is
        "Video".
    data_keys
        Optional. The data key(s) in ts_video to plot. Default is [], which
        means that every data of the TimeSeries is plotted.
    event_source
        Optional. Defines which TimeSeries to read the event from. Can be::
            - "merge" (default): The editor includes the events from both
              video_ts and data_ts
            - "video": The editor includes only the events from video_ts
            - "data_ts": The editor includes only the events from data_ts
    in_place
        Optional. True to modify the original TimeSeries, False to return a
        copy.

    Returns
    -------
    list[TimeSeries]
        The list [video_ts, data_ts] which are now synchronized and which
        share the events.

    """
    warn_once(
        "Function ui_edit_events is experimental and will probably change in "
        "the future."
    )

    if not in_place:
        ts_video = ts_video.copy()
        ts_data = ts_data.copy()

    video = Video(
        ts_video=ts_video,
        ts_data=ts_data,
        video_key=video_key,
        data_keys=data_keys,
        event_source=event_source,
    )

    while video._closed is False:
        plt.pause(0.1)

    return [ts_video, ts_data]


class Video:
    """
    Launch an interactive video player to edit events (Work in progress).

    Warning
    -------
    This feature is experimental and will probably change in the future.

    Attributes
    ----------
    ts_video : TimeSeries
        The TimeSeries that contains the video, normally obtained using
        the read_video() function.
    ts_data : TimeSeries
        Optional. TimeSeries.
    video_key : str
        Optional. The data key in ts_video that contains the video. Default is
        "Video".
    data_keys : str | list[str]
        Optional. The data key(s) in ts_video to plot. Default is [], which
        means that every data of the TimeSeries is plotted.
    event_source: str
        Optional. Defines which TimeSeries to read the event from. Can be::
            - "merge" (default): The editor includes the events from both
              video_ts and data_ts
            - "video": The editor includes only the events from video_ts
            - "data_ts": The editor includes only the events from data_ts

    """

    def __init__(
        self,
        ts_video: TimeSeries,
        ts_data: TimeSeries | None,
        *,
        video_key: str = "Video",
        data_keys: str | list[str] = [],
        event_source: str = "merge",
    ):

        self._closed = False

        self._video_key = video_key
        self._data_keys = data_keys

        # Create the Video TimeSeries
        self._ts_video = ts_video

        # Create the Data TimeSeries
        if ts_data is None:
            self._ts_data = TimeSeries(
                time=ts_video.time, events=ts_video.events
            )
        else:
            self._ts_data = ts_data

        # Select which events to keep or merge
        if event_source == "merge":
            for event in self._ts_video.events:
                self._ts_data.add_event(
                    event.time, event.name, unique=True, in_place=True
                )
            for event in self._ts_data.events:
                self._ts_video.add_event(
                    event.time, event.name, unique=True, in_place=True
                )
        elif event_source == "data":
            self._ts_video.events = []
            for event in self._ts_data.events:
                self._ts_video.add_event(event.time, event.name, in_place=True)
        elif event_source == "video":
            self._ts_data.events = []
            for event in self._ts_video.events:
                self._ts_data.add_event(event.time, event.name, in_place=True)
        else:
            raise ValueError(
                "event_source must be 'merge', 'data' or 'video'."
            )

        # Extract information from the video
        self._video_length = self._ts_video.data[video_key].shape[0]
        self._video_height = self._ts_video.data[video_key].shape[1]
        self._video_width = self._ts_video.data[video_key].shape[2]

        # Ensure we work with bytes, in case someone did operations on it and
        # now we have floats.
        self._ts_video.data[video_key] = np.clip(
            self._ts_video.data[video_key], 0, 255
        ).astype(np.uint8)

        self._create_gui()

        # Public properties
        self.current_index = 0

        # Last things
        self._refresh()

    # Properties
    @property
    def current_index(self) -> int:
        """Read/write current_index."""
        return self._current_index

    @current_index.setter
    def current_index(self, value: int):
        """Set current_index value."""
        self._current_index = value % self._video_length
        self._refresh()

    @property
    def current_time(self) -> float:
        """Read/write current_time."""
        return self._ts_video.time[self._current_index]

    @current_time.setter
    def current_time(self, value: float):
        """Set current_time value."""
        index = int(np.argmin(np.abs(self._ts_video.time - value)))
        self.current_index = index

    def _create_gui(self):
        """Create all the GUI elements."""
        # Remove default keymaps for left and right
        if "left" in plt.rcParams["keymap.back"]:
            plt.rcParams["keymap.back"].remove("left")
        if "right" in plt.rcParams["keymap.forward"]:
            plt.rcParams["keymap.forward"].remove("right")

        # Create the figure
        # fmt: off
        self._figure, self._axes = plt.subplot_mosaic(
            [
                ["LblNextIndex",  "LblNextIndex", "Video",    "LblEvent",     "LblEvent"],
                ["LblNextIndex",  "LblNextIndex", "Video",    "LblEvent",     "LblEvent"],
                ["BtnPrevIndex",  "BtnNextIndex", "Video",    "TxtEventName", "TxtEventName"],
                ["LblNextEvent",  "LblNextEvent", "Video",    "BtnAddEvent",  "BtnRemoveEvent"],
                ["LblNextEvent",  "LblNextEvent", "Video",    "RSpace1",      "RSpace1"],
                ["BtnPrevEvent",  "BtnNextEvent", "Video",    "RSpace1",      "RSpace1"],
                ["LSpace",        "LSpace",       "Video",    "LblZero",      "LblZero"],
                ["LSpace",        "LSpace",       "Video",    "LblZero",      "LblZero"],
                ["LSpace",        "LSpace",       "Video",    "BtnZeroVideo", "BtnZeroVideo"],
                ["LSpace",        "LSpace",       "Video",    "BtnZeroData",  "BtnZeroData"],
                ["BtnPlay",       "BtnPlay",      "Video",    "RSpace2",      "RSpace2"],
                ["Space1",        "Space1",       "Space1",   "Space1",       "Space1"],
                ["Data",          "Data",         "Data",     "Data",         "Data"],
            ],
            width_ratios=[1, 1, 20, 1, 1],
            height_ratios=[1,1, 1, 1,1, 1, 1, 1, 1, 1, 1, 1.5, 6],
        )
        # fmt: on

        # Remove frames and ticks for all for now
        for key in self._axes:
            self._axes[key].set_frame_on(False)
            self._axes[key].set_xticks([])
            self._axes[key].set_yticks([])

        # Create the video part
        self._axes["Video"].set_frame_on(True)
        self._axes_image = self._axes["Video"].imshow(
            self._ts_video.data[self._video_key][0],
            interpolation="none",
            resample=False,
        )

        # Create the data part
        self._axes_data = self._axes["Data"]
        self._create_refresh_data()

        # Create a state like in the Player
        self._state = {"SystemTimeOnLastUpdate": time.time()}
        self._running = False

        # Create the anim timer
        self._anim = animation.FuncAnimation(
            self._figure,
            self._on_timer,  # type: ignore
            interval=33,
            cache_frame_data=False,
        )  # 30 ips

        # Create the GUI elements

        self._axes["LblNextIndex"].text(
            0.5,
            0,
            "Next Index\n(←/→/Scroll)",
            horizontalalignment="center",
            verticalalignment="bottom",
        )
        self._axes["BtnPrevIndex"].set_frame_on(True)
        self._btn_prev_index = widgets.Button(self._axes["BtnPrevIndex"], "◀︎")
        self._btn_prev_index.on_clicked(self._on_previous_index)
        self._axes["BtnNextIndex"].set_frame_on(True)
        self._btn_next_index = widgets.Button(self._axes["BtnNextIndex"], "▶︎")
        self._btn_next_index.on_clicked(self._on_next_index)

        self._axes["LblNextEvent"].text(
            0.5,
            0,
            "Next Event\n(Shift + ←/→)",
            horizontalalignment="center",
            verticalalignment="bottom",
        )
        self._axes["BtnPrevEvent"].set_frame_on(True)
        self._btn_prev_event = widgets.Button(self._axes["BtnPrevEvent"], "◀︎◀︎")
        self._btn_prev_event.on_clicked(self._on_previous_event)
        self._axes["BtnNextEvent"].set_frame_on(True)
        self._btn_next_event = widgets.Button(self._axes["BtnNextEvent"], "▶︎▶︎")
        self._btn_next_event.on_clicked(self._on_next_event)

        self._axes["LblZero"].text(
            0.5,
            0,
            "Sync (zero)",
            horizontalalignment="center",
            verticalalignment="bottom",
        )
        self._axes["BtnZeroVideo"].set_frame_on(True)
        self._btn_zero_video = widgets.Button(
            self._axes["BtnZeroVideo"], "Video (z)"
        )
        self._btn_zero_video.on_clicked(self._on_zero_video)
        self._axes["BtnZeroData"].set_frame_on(True)
        self._btn_zero_data = widgets.Button(
            self._axes["BtnZeroData"], "Data (Shift + z)"
        )
        self._btn_zero_data.on_clicked(self._on_zero_data)

        self._axes["LblEvent"].text(
            0.5,
            0,
            "Add/Remove",
            horizontalalignment="center",
            verticalalignment="bottom",
        )
        self._axes["TxtEventName"].set_frame_on(True)
        self._txt_event_name = widgets.TextBox(
            self._axes["TxtEventName"], label="", initial="event", color="w"
        )
        self._axes["BtnAddEvent"].set_frame_on(True)
        self._btn_add_event = widgets.Button(self._axes["BtnAddEvent"], "+")
        self._btn_add_event.on_clicked(self._on_add_event)
        self._axes["BtnRemoveEvent"].set_frame_on(True)
        self._btn_remove_event = widgets.Button(
            self._axes["BtnRemoveEvent"], "−"
        )
        self._btn_remove_event.on_clicked(self._on_remove_event)

        # Connect callbacks to the figure
        self._figure.canvas.mpl_connect("key_press_event", self._on_key)
        # self._figure.canvas.mpl_connect("key_release_event", self._on_release)
        self._figure.canvas.mpl_connect("scroll_event", self._on_scroll)
        self._figure.canvas.mpl_connect("pick_event", self._on_pick)
        self._figure.canvas.mpl_connect("close_event", self._on_close)

        # Give a decent size to the figure
        self._figure.set_size_inches(15, 8)

    # %% Callbacks

    def _on_next_index(self, event):
        self.current_index += 1

    def _on_previous_index(self, event):
        self.current_index -= 1

    def _on_next_event(self, event):
        for i, one_event in enumerate(self._ts_video.events):
            if one_event.time > self.current_time:
                self.current_time = one_event.time
                break
        self._create_refresh_data()
        self._refresh()

    def _on_previous_event(self, event):
        for i, one_event in enumerate(self._ts_video.events[-1::-1]):
            if one_event.time < self.current_time:
                self.current_time = one_event.time
                break
        self._create_refresh_data()
        self._refresh()

    def _on_zero_video(self, event):
        shift = -self.current_time
        self._ts_video.shift(shift, in_place=True)
        self._create_refresh_data()
        self.current_time = 0
        self._refresh()

    def _on_zero_data(self, event):
        shift = -self.current_time
        self._ts_data.shift(shift, in_place=True)
        self._create_refresh_data()
        self.current_time = 0
        self._refresh()

    def _on_add_event(self, event):
        self._ts_data.add_event(
            self.current_time,
            self._txt_event_name.text,
            in_place=True,
            unique=True,
        )
        self._ts_video.add_event(
            self.current_time,
            self._txt_event_name.text,
            in_place=True,
            unique=True,
        )
        self._create_refresh_data()
        self._refresh()

    def _on_remove_event(self, event):
        for i, one_event in enumerate(self._ts_video.events):
            if (
                np.isclose(one_event.time, self.current_time)
                and one_event.name == self._txt_event_name.text
            ):
                self._ts_data.events.pop(i)
                self._ts_video.events.pop(i)
                break
        self._create_refresh_data()
        self._refresh()

    def _on_key(self, event):
        if event.key == "right":
            self._on_next_index(event)

        elif event.key == "left":
            self._on_previous_index(event)

        elif event.key == "shift+right":
            self._on_next_event(event)

        elif event.key == "shift+left":
            self._on_previous_event(event)

        elif event.key == " ":
            if self._running:
                self._running = False
                self._anim.event_source.stop()
            else:
                self._state["SystemTimeOnLastUpdate"] = time.time()
                self._running = True
                self._anim.event_source.start()

        elif event.key == "+":
            self._on_add_event(event)

        elif event.key == "-":
            self._on_remove_event(event)

        elif event.key == "z":
            self._on_zero_video(event)

        elif event.key == "Z":
            self._on_zero_data(event)

    def _on_pick(self, event):
        if event.mouseevent.button == 1:
            index = event.ind[0]
            self.current_index = index

    def _on_scroll(self, event):  # pragma: no cover
        if event.button == "up":
            self.current_index -= 1
        elif event.button == "down":
            self.current_index += 1

    def _on_timer(self, _) -> None:  # pragma: no cover
        """Implement callback for the animation timer object."""
        if self._running is True:
            # We check self._running because we can enter this callback
            # even if the animation has been deactivated. This is because the
            # recommended way to deactivate a timer is to unreference it,
            # however the garbage collector may take time deleting the timer
            # and we will end up with still a few timer callbacks. Checking
            # self._running makes sure that we effectively stop.
            current_system_time = time.time()
            current_index = self.current_index
            self.current_time += 1.0 * (
                time.time() - self._state["SystemTimeOnLastUpdate"]  # type: ignore
            )
            # Type ignored because mypy considers
            # self._state["SystemTimeOnLastUpdate"] as an "object"
            if current_index == self.current_index:
                # The time wasn't enough to advance a frame. Articifically
                # advance a frame.
                self.current_index += 1
            self._state["SystemTimeOnLastUpdate"] = current_system_time
        else:
            self._anim.event_source.stop()

    def _refresh(self):
        self._axes_image.set_data(
            self._ts_video.data["Video"][self._current_index]
        )
        #        self._axes.set_title(
        #            f"{self._current_index} ({self.current_time:.2f} s.)"
        #        )

        # Update line
        cursor_data = self._axes_index_line.get_data()
        self._axes_index_line.set_data(
            [
                self._ts_video.time[self._current_index],
                self._ts_video.time[self._current_index],
            ],
            [cursor_data[1][0], cursor_data[1][1]],
        )

        # Update cursor
        cursor_data = self._axes_index_cursor.get_data()
        self._axes_index_cursor.set_data(
            [
                self._ts_video.time[self._current_index],
            ],
            [cursor_data[1][0]],
        )

        self._figure.canvas.draw_idle()

    def _create_refresh_data(self):
        plt.sca(self._axes_data)
        plt.cla()

        # Plot data
        self._ts_data.plot(self._data_keys, legend=False)
        plt.xlabel("")
        plt.ylabel("")

        # Plot red line
        current_axes = plt.axis()
        self._axes_index_line = plt.plot(
            [0, 0], [current_axes[2], current_axes[3]], "r"
        )[0]

        # Plot clickable timeline
        plt.plot(
            self._ts_video.time,
            (0 * self._ts_video.time)
            + current_axes[2]
            - 0.1 * (current_axes[3] - current_axes[2]),
            ".",
            color=(0.8, 0.8, 0.8),
            picker=True,
            pickradius=2,
        )

        # Plot red cursor
        self._axes_index_cursor = plt.plot(
            [0],
            [
                current_axes[2] - 0.1 * (current_axes[3] - current_axes[2]),
            ],
            "dr",
            markersize=8,
        )[0]

        plt.axis(
            [
                min(self._ts_video.time[0], self._ts_data.time[0]),
                max(self._ts_video.time[-1], self._ts_data.time[-1]),
                current_axes[2] - 0.15 * (current_axes[3] - current_axes[2]),
                current_axes[3],
            ]
        )

    def _on_close(self, event):
        self._closed = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020-2025 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import kineticstoolkit as ktk
import numpy as np


TEST_CALIBRATION_MATRIX = {
    "gains": np.array([-0.106, 0.106, 0.094, 0.022, -0.022, 0.0234999]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}


def test_read_smartwheel():
    """Test that read_file works similarly for SW's csv and txt files."""
    kinetics_csv = ktk.ext.pushrimkinetics.read_smartwheel(
        ktk.ext.root_folder
        + "/data/pushrimkinetics_propulsion.csv"
    )

    kinetics_txt = ktk.ext.pushrimkinetics.read_smartwheel(
        ktk.ext.root_folder
        + "/data/pushrimkinetics_propulsion.txt"
    )

    smaller = min(kinetics_csv.time.shape[0], kinetics_txt.time.shape[0])

    assert np.allclose(
        kinetics_csv.data["Channels"][0:smaller],
        kinetics_txt.data["Channels"][0:smaller],
    )
    assert (
        np.std(
            kinetics_csv.data["Angle"][0:smaller]
            - kinetics_txt.data["Angle"][0:smaller]
        )
        < 1e-4
    )


def test_remove_offsets():
    """Test that remove_offsets works with and without a baseline."""
    kinetics = ktk.ext.pushrimkinetics.read_smartwheel(
        ktk.ext.root_folder
        + "/data/pushrimkinetics_offsets_propulsion.csv"
    )

    baseline = ktk.ext.pushrimkinetics.read_smartwheel(
        ktk.ext.root_folder
        + "/data/pushrimkinetics_offsets_baseline.csv"
    )

    no_offsets1 = ktk.ext.pushrimkinetics.remove_offsets(kinetics)
    no_offsets2 = ktk.ext.pushrimkinetics.remove_offsets(kinetics, baseline)

    # Assert that all force differences are within 1 N
    assert np.all(
        np.abs(no_offsets1.data["Forces"] - no_offsets2.data["Forces"]) < 1
    )

    # Assert that all moment differences are within 0.1 Nm
    assert np.all(
        np.abs(no_offsets1.data["Moments"] - no_offsets2.data["Moments"]) < 0.1
    )


def test_apply_calibration():
    """Test that force calculation is similar to precalculated forces."""
    kinetics = ktk.ext.pushrimkinetics.read_smartwheel(
        ktk.ext.root_folder
        + "/data/pushrimkinetics_offsets_propulsion.csv"
    )

    test = kinetics.copy()
    test = ktk.ext.pushrimkinetics.apply_calibration(
        test,
        **TEST_CALIBRATION_MATRIX,
        reference_frame="hub",
    )

    assert np.allclose(
        np.mean(np.abs(test.data["Forces"]), axis=0),
        np.mean(np.abs(kinetics.data["Forces"]), axis=0),
        atol=2,
    )
    assert np.allclose(
        np.std(np.abs(test.data["Forces"]), axis=0),
        np.std(np.abs(kinetics.data["Forces"]), axis=0),
        atol=2,
    )
    assert np.allclose(
        np.mean(np.abs(test.data["Moments"]), axis=0),
        np.mean(np.abs(kinetics.data["Moments"]), axis=0),
        atol=2,
    )
    assert np.allclose(
        np.std(np.abs(test.data["Moments"]), axis=0),
        np.std(np.abs(kinetics.data["Moments"]), axis=0),
        atol=2,
    )


def test_calculate_velocity_power():
    """No-regression test for calculate_velocity and calculate_power."""
    kinetics = ktk.ext.pushrimkinetics.read_smartwheel(
        ktk.ext.root_folder
        + "/data/pushrimkinetics_offsets_propulsion.csv"
    )

    kinetics = ktk.ext.pushrimkinetics.calculate_velocity(kinetics)
    assert np.allclose(
        [
            np.mean(kinetics.data["Velocity"]),
            np.std(kinetics.data["Velocity"]),
        ],
        [2.875997177730561, 0.8584197191949383],
    )

    kinetics = ktk.ext.pushrimkinetics.calculate_power(kinetics)
    assert np.allclose(
        kinetics.data["Velocity"] * kinetics.data["Moments"][:, 2],
        kinetics.data["Power"],
    )


if __name__ == "__main__":
    import pytest

    pytest.main([__file__])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides functions to process kinetic data from instrumented wheelchair wheels.
"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020-2025 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import kineticstoolkit as ktk
import numpy as np
import pandas as pd
import struct  # to unpack binary data from SmartWheels' txt files


def __dir__():
    return [
        "read_smartwheel",
        "remove_offsets",
        "apply_calibration",
        "calculate_velocity",
        "calculate_power",
    ]


def read_smartwheel(filename: str) -> ktk.TimeSeries:
    """
    Read a file generated by the SmartWheel instrumented wheel.

    Parameters
    ----------
    filename
        Name of the file to open

    Returns
    -------
    TimeSeries
        The file contents.

    """
    if filename.lower().endswith(".csv"):

        dataframe = pd.read_csv(filename, delimiter=None, header=None)
        if dataframe.shape[1] == 1:  # Retry with ; as separator
            dataframe = pd.read_csv(filename, delimiter=";", header=None)

        data = dataframe.to_numpy()
        index = data[:, 1]
        time = np.arange(0, len(index)) / 240
        channels = data[:, 6:12]
        forces = data[:, 18:21]
        moments = data[:, 21:24]
        angle_deg = data[:, 3]
        angle_rad = np.unwrap(np.deg2rad(angle_deg))

        ts = ktk.TimeSeries(time=time)

        ts.data["Index"] = index
        ts.data["Channels"] = channels
        ts.data["Forces"] = np.block([[forces, np.zeros((len(index), 1))]])
        ts.data["Moments"] = np.block([[moments, np.zeros((len(index), 1))]])
        ts.data["Angle"] = angle_rad

        ts = ts.add_data_info("Channels", "Unit", "raw")
        ts = ts.add_data_info("Forces", "Unit", "N")
        ts = ts.add_data_info("Moments", "Unit", "Nm")
        ts = ts.add_data_info("Angle", "Unit", "rad")

    elif filename.lower().endswith(".txt"):

        data = {
            "ch1": [],
            "ch2": [],
            "ch3": [],
            "ch4": [],
            "ch5": [],
            "ch6": [],
            "angle_ticks": [],
        }

        length = 0
        with open(filename, "rb") as fid:

            while True:
                try:
                    _ = fid.read(2)
                    data["ch1"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch2"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch3"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch4"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch5"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch6"].append(struct.unpack("h", fid.read(2))[0])
                    data["angle_ticks"].append(
                        struct.unpack("i", fid.read(4))[0]
                    )
                    _ = fid.read(8)
                    length += 1
                except Exception:
                    break

        ch1 = np.array(data["ch1"][1:length])  # Remove 1st sample to be
        ch2 = np.array(data["ch2"][1:length])  # consistent with CSV file
        ch3 = np.array(data["ch3"][1:length])
        ch4 = np.array(data["ch4"][1:length])
        ch5 = np.array(data["ch5"][1:length])
        ch6 = np.array(data["ch6"][1:length])
        angle_ticks = np.array(data["angle_ticks"][1:length])

        # Keep only 12 least significant bytes
        ch1 = np.mod(ch1, 2**12)
        ch2 = np.mod(ch2, 2**12)
        ch3 = np.mod(ch3, 2**12)
        ch4 = np.mod(ch4, 2**12)
        ch5 = np.mod(ch5, 2**12)
        ch6 = np.mod(ch6, 2**12)

        # Convert angle in radian
        angle = angle_ticks / 4096 * 2 * np.pi

        ts = ktk.TimeSeries(
            time=np.linspace(0, (length - 1) / 240, length - 1)
        )
        ts.data["Channels"] = np.concatenate(
            [
                ch1[:, np.newaxis],
                ch2[:, np.newaxis],
                ch3[:, np.newaxis],
                ch4[:, np.newaxis],
                ch5[:, np.newaxis],
                ch6[:, np.newaxis],
            ],
            axis=1,
        )
        ts.data["Angle"] = angle

        ts = ts.add_data_info("Channels", "Unit", "raw")
        ts = ts.add_data_info("Angle", "Unit", "rad")

    else:
        raise ValueError("Unknown file format.")

    return ts


def _find_recovery_indices(Mz: np.ndarray) -> np.ndarray:
    """
    Find recovery indices based on a vector of propulsion moments.

    This function analyzes the Mz moments to find which data correspond to
    pushes and which data correspond to recoveries. The method is very
    conservative on what could be considered as a recovery, so that every
    index returned by this function is almost certain to correspond to a
    recovery. This function is used by `pushrimkinetics.remove_sinusoids`
    to identify the instants with no hand contact. It should not be used to
    isolate the push and recovery phases (use `ktk.cycles.detect_cycles()`
    instead).

    Parameters
    ----------
    Mz
        Array that contains the propulsion moments in Nm.

    Returns
    -------
    np.ndarray
        Array of bools where each True represents recovery.

    """
    Mz = Mz.copy()

    threshold = 2.24  # (Nm): max tolerance for the remaining values.

    while np.nanmax(Mz) - np.nanmin(Mz) > threshold:

        # Remove 1% of data that are the farthest to the median:

        # Sort data
        index_to_remove = np.argsort(np.abs(Mz - np.nanmedian(Mz)))
        sorted_Mz = Mz[index_to_remove]
        index_to_remove = index_to_remove[~np.isnan(sorted_Mz)]

        # Remove the 1% upper.
        index_to_remove = index_to_remove[
            int(0.99 * len(index_to_remove)) - 1 :
        ]

        # Assign nan to these data
        Mz[index_to_remove] = np.nan

    index = ~np.isnan(Mz)

    return index


def remove_offsets(
    ts: ktk.TimeSeries, baseline_kinetics: ktk.TimeSeries | None = None
) -> ktk.TimeSeries:
    """
    Remove dynamic offsets in forces and moments.

    Parameters
    ----------
    ts
        TimeSeries that contains at least Forces, Moments and Angle data.
    baseline_kinetics
        Optional. TimeSeries that contains at least Forces and Moments data.
        This TimeSeries contains a baseline trial, where the wheelchair must be
        pushed by an operator and where no external force must be applied on
        the pushrims. If no baseline is provided, the baseline is calculated
        based on a detection of recoveries in the supplied ts
        TimeSeries.

    Returns
    -------
    TimeSeries
        A copy of the input TimeSeries with removed sinusoids in Forces and
        Moments data.

    References
    ----------
    F. Chénier, R. Aissaoui, C. Gauthier, and D. H. Gagnon,
    "Wheelchair pushrim kinetics measurement: A method to cancel
    inaccuracies due to pushrim weight and wheel camber," Medical
    Engineering and Physics, vol. 40, pp. 75--86, 2017.

    """
    ts_out = ts.copy()

    if baseline_kinetics is None:
        # Create baseline kinetics.
        recovery_index = _find_recovery_indices(ts.data["Moments"][:, 2])
        f_ofs = np.hstack(
            (
                ts.data["Forces"][recovery_index, 0:3],
                ts.data["Moments"][recovery_index, 0:3],
            )
        )
        theta_baseline = ts.data["Angle"][recovery_index]

    else:
        # Use baseline kinetics.
        f_ofs = np.hstack(
            (
                baseline_kinetics.data["Forces"][:, 0:3],
                baseline_kinetics.data["Moments"][:, 0:3],
            )
        )
        theta_baseline = baseline_kinetics.data["Angle"][:]

    # Do the regression
    theta_baseline = theta_baseline[:, np.newaxis]
    q = np.hstack(
        (
            np.sin(theta_baseline),
            np.cos(theta_baseline),
            np.ones((len(theta_baseline), 1)),
        )
    )
    A = np.linalg.lstsq(q, f_ofs, rcond=None)
    A = A[0]

    # Apply the regression to forces and moments
    theta = ts.data["Angle"]
    theta = theta[:, np.newaxis]

    f = np.hstack((ts.data["Forces"][:, 0:3], ts.data["Moments"][:, 0:3]))

    q = np.hstack((np.sin(theta), np.cos(theta), np.ones((len(theta), 1))))

    f = f - q @ A

    # Make the output timeseries
    ts_out.data["Forces"] = np.zeros((f.shape[0], 4))
    ts_out.data["Forces"][:, 0:3] = f[:, 0:3]
    ts_out.data["Forces"][:, 3] = 0
    ts_out.data["Moments"] = np.zeros((f.shape[0], 4))
    ts_out.data["Moments"][:, 0:3] = f[:, 3:6]
    ts_out.data["Moments"][:, 3] = 0
    ts_out.add_data_info("Forces", "Unit", "N", in_place=True, overwrite=True)
    ts_out.add_data_info(
        "Moments", "Unit", "Nm", in_place=True, overwrite=True
    )

    return ts_out


def apply_calibration(
    ts: ktk.TimeSeries,
    gains: np.ndarray | str,
    offsets: np.ndarray = np.zeros((6)),
    *,
    transducer: str = "force_cell",
    reference_frame: str = "wheel"
) -> ktk.TimeSeries:
    """
    Calculate pushrim forces and moments based on raw channel values.

    For standard force cells (with each channel being a raw value
    corresponding to Fx, Fy, Fz, Mx, My, Mz, respectively), calculates
    the forces and moments using a sensitivity matrix (gains) and an
    offset vector (offsets).

    For SmartWheel, calculates the forces and moments using a gain
    vector (gains) and an offset vector (offsets).

        [Fx, Fy, Fz, Mx, My, Mz] = gains @ channels + offsets

    Parameters
    ----------
    ts
        Input TimeSeries that must contain a 'Channels' key in its data dict.
    gains
        6x6 gain matrix (force_cell) or gain vector of length 6 (smartwheel).
    offsets
        Optional. Offset vector of length 6.
    transducer
        Optional. 'force_cell' or 'smartwheel'.
    reference_frame
        Optional. 'wheel' or 'hub'. 'wheel' to report the forces and moments
        into the local wheel's reference frame; 'hub' to compensate for the
        wheel rotation and match the reference frame used by the SmartWheel:
        x anteroposterior, y in the wheel plane, upward for non-camberred
        wheels, and z perpendicular to the wheel plane, outward.

    Returns
    -------
    TimeSeries
        A copy of the input TimeSeries with added Forces and Moments data.

    """
    # Calculate the forces and moments and add to the output
    if transducer == "smartwheel":

        # Calculate the rotation angle to apply to the calculated ts
        if reference_frame == "wheel":
            theta = np.array([0.0])
        elif reference_frame == "hub":
            theta = ts.data["Angle"]
        else:
            raise ValueError("reference_frame must be 'wheel' or 'hub'")

        # Extract channels and angle
        ch = ts.data["Channels"] - 2048

        # Calculate the forces and moments
        Fx = (
            gains[0]
            * (
                ch[:, 0] * np.sin(theta)
                + ch[:, 2] * np.sin(theta + 2 * np.pi / 3)
                + ch[:, 4] * np.sin(theta + 4 * np.pi / 3)
            )
            + offsets[0]
        )

        Fy = (
            gains[1]
            * (
                ch[:, 0] * np.cos(theta)
                + ch[:, 2] * np.cos(theta + 2 * np.pi / 3)
                + ch[:, 4] * np.cos(theta + 4 * np.pi / 3)
            )
            + offsets[1]
        )

        Fz = gains[2] * (ch[:, 1] + ch[:, 3] + ch[:, 5]) + offsets[2]

        Mx = (
            gains[3]
            * (
                ch[:, 1] * np.sin(theta)
                + ch[:, 3] * np.sin(theta + 2 * np.pi / 3)
                + ch[:, 5] * np.sin(theta + 4 * np.pi / 3)
            )
            + offsets[3]
        )

        My = (
            gains[4]
            * (
                ch[:, 1] * np.cos(theta)
                + ch[:, 3] * np.cos(theta + 2 * np.pi / 3)
                + ch[:, 5] * np.cos(theta + 4 * np.pi / 3)
            )
            + offsets[4]
        )

        Mz = gains[5] * (ch[:, 0] + ch[:, 2] + ch[:, 4]) + offsets[5]
        forces_moments = np.block(
            [
                Fx[:, np.newaxis],
                Fy[:, np.newaxis],
                Fz[:, np.newaxis],
                Mx[:, np.newaxis],
                My[:, np.newaxis],
                Mz[:, np.newaxis],
            ]
        )

    elif transducer == "force_cell":

        # Calculate the rotation angle to apply to the calculated ts
        if reference_frame == "wheel":
            theta = np.array([0.0])
        elif reference_frame == "hub":
            raise NotImplementedError(
                "hub reference_frame not implemented yet"
                "for force_cell transducers."
            )
        else:
            raise ValueError("reference_frame must be 'wheel' or 'hub'")

        n_frames = ts.data["Channels"].shape[0]

        forces_moments = np.empty((n_frames, 6))
        for i_frame in range(n_frames):
            forces_moments[i_frame] = (
                gains @ ts.data["Channels"][i_frame] + offsets
            )

    # Format these data in the output timeseries
    ts_out = ts.copy()

    ts_out.data["Forces"] = np.concatenate(
        [forces_moments[:, 0:3], np.zeros((forces_moments.shape[0], 1))],
        axis=1,
    )
    ts_out.add_data_info("Forces", "Unit", "N", in_place=True, overwrite=True)

    ts_out.data["Moments"] = np.concatenate(
        [forces_moments[:, 3:6], np.zeros((forces_moments.shape[0], 1))],
        axis=1,
    )
    ts_out.add_data_info(
        "Moments", "Unit", "Nm", in_place=True, overwrite=True
    )

    return ts_out


def calculate_velocity(ts: ktk.TimeSeries) -> ktk.TimeSeries:
    """
    Calculate velocity based on wheel angle.

    The velocity is calculated by deriving the angle using a 2nd order
    Savitzky-Golay filter of length 21. This filter has been experimentally
    validated to maximize the signal-to-noise ratio for a SmartWheel recording
    at 240 Hz. This function may change signature in the future to include
    other filtering options. To manually get the velocity using a custom
    filter, please see the ``ktk.filters`` module.

    Parameters
    ----------
    ts
        TimeSeries that contains at least the data key 'Angle'.

    Returns
    -------
    TimeSeries
         A copy of the input TimeSeries with the added 'Velocity' data.

    """
    tsangle = ktk.TimeSeries()
    tsangle.time = ts.time
    tsangle.data["Angle"] = ts.data["Angle"]
    tsvelocity = ktk.filters.savgol(
        tsangle, window_length=21, poly_order=2, deriv=1
    )
    tsout = ts.copy()
    tsout.data["Velocity"] = tsvelocity.data["Angle"]
    try:
        tsout.add_data_info(
            "Velocity",
            "Unit",
            tsout.data_info["Angle"]["Unit"] + "/s",
            in_place=True,
        )
    except KeyError:
        pass
    return tsout


def calculate_power(ts: ktk.TimeSeries) -> ktk.TimeSeries:
    """
    Calculate power based on wheel velocity and moment.

    Parameters
    ----------
    tsin
        TimeSeries that contains at least the data keys 'Velocity' and
        'Moments'. The units must be consistent (e.g., rad/s and Nm)

    Returns
    -------
    TimeSeries
        A copy of the TimeSeries with the added 'Power' data.

    """
    tsout = ts.copy()
    tsout.data["Power"] = ts.data["Velocity"] * ts.data["Moments"][:, 2]
    tsout.add_data_info("Power", "Unit", "W", in_place=True)
    return tsout



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("kineticstoolkit_extensions/VERSION", "r") as fh:
    version = fh.read()


setuptools.setup(
    name="kineticstoolkit_extensions",
    version=version,
    description="Extensions and unstable development modules for Kinetics Toolkit.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://kineticstoolkit.uqam.ca",
    author="Félix Chénier",
    author_email="chenier.felix@uqam.ca",
    license="Apache",
    license_files=["LICENSE.txt", "NOTICE.txt"],
    packages=setuptools.find_packages(),
    package_data={
        "kineticstoolkit_extensions": ["VERSION"],
    },
    project_urls={
        "Documentation": "https://kineticstoolkit.uqam.ca",
        "Source": "https://github.com/kineticstoolkit/kineticstoolkit_extensions",
        "Tracker": "https://github.com/kineticstoolkit/kineticstoolkit/issues",
    },
    install_requires=["kineticstoolkit"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.10",
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Kinetics Toolkit Extensions."""
__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020-2025 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"

import os

import kineticstoolkit_extensions.n3d as n3d
import kineticstoolkit_extensions.pushrimkinetics as pushrimkinetics
import kineticstoolkit_extensions.video as video

def __dir__():
    return ["n3d", "pushrimkinetics"]


root_folder = os.path.dirname(os.path.dirname(__file__))
